/* dump.c */

#include "fuzz.h"
#include "symbol.h"
#include "absyn.h"

PRIVATE void dump(tree t, int indent);

PRIVATE void in_dump(tree t, int indent)
{
     printf("\n%*s", 2*indent, "");
     dump(t, indent);
}

PRIVATE void dump_list(tree t, int indent)
{
     if (t == nil)
	  printf(" ()");
     else
	  in_dump(t, indent);
}

PRIVATE void dump(tree t, int indent)
{
     int i; tree u;

     if (t == nil) {
	  printf("()");
	  return;
     }

     switch (t->x_kind) {
     case CONS:
	  printf("(LIST");
	  for (u = t; u != nil; u = cdr(u))
	       in_dump(car(u), indent+1);
	  printf(")");
	  break;

     case WORD:
	  printf("\"%s\"", ((tok) t)->l_chars);
	  break;

     case IDENT:
	  printf("(NAME \"%s\" \"%s\")", 
		 ((sym) t)->s_basename->l_chars,
		 ((sym) t)->s_decor->l_chars);
	  break;

     case REF:
	  printf("(REF %d ", t->x_loc);
	  dump(t->x_tag, 0);
	  dump_list(t->x_params, indent+1);
	  printf(")");
	  break;

     case SREF:
	  printf("(SREF %d ", t->x_loc);
	  dump(t->x_sref_tag, 0);
	  printf(" ");
	  dump(t->x_sref_decor, 0);
	  dump_list(t->x_sref_params, indent+1);
	  dump_list(t->x_sref_renames, indent+1);
	  printf(")");
	  break;

     case NUMBER:
	  printf("(NUMBER %d \"%s\")", t->x_loc, (char *) t->x_number);
	  break;

     default:
	  printf("(%s %d", name_of_tag(t->x_kind), t->x_loc);
	  for (i = 0; i < t->x_arity; i++)
	       in_dump(t->x_slot[i], indent+1);
	  printf(")");
     }
}

PUBLIC void dump_tree(tree t)
{
     dump(t, 0);
     printf("\n");
}
