/* 
 * The fuzz type-checker for Z
 * File: frame.c
 * $Revision: 3.0 $
 * Copyright (C) J. M. Spivey 1989, 1991
 */

#include "fuzz.h"

PRIVATE struct {
     frame b_chain;
     bool b_mark;
} stack[MAX_LEVEL];
PRIVATE int level = -1;

/* begin_frame -- start recording frames */
PUBLIC void begin_frame(void)
{
     if (level >= MAX_LEVEL)
	  panic("begin_frame -- Nesting too deep");

     level++;
     stack[level].b_chain = NULL;
     stack[level].b_mark = FALSE;
}

/* end_frame -- check all type variables bound */
PUBLIC void end_frame(void)
{
     frame p;
     int j;
     
     for (p = stack[level].b_chain; p != NULL; p = p->f_next) {
	  int failed = FALSE;

	  for (j = 0; j < p->f_nvars; j++) {
	       if (p->f_var[j] == NULL) {
		    failed = TRUE;
		    p->f_var[j] = err_type;
	       }
	  }

	  if (failed && ! stack[level].b_mark) {
	       tc_error(p->f_cxt->x_loc,
			"Implicit parameters not completely determined");
	       tc_e_etc("Expression: %z", p->f_cxt);
	       tc_e_end();
	  }
     }

     level--;
}
     
/* mark_error -- don't complain about unbound variables */
PUBLIC void mark_error(void)
{
     stack[level].b_mark = TRUE;
}

/* mk_frame -- allocate a new frame */
PUBLIC frame mk_frame(int nvars)
{
     if (nvars == 0)
	  return arid;
     else {
	  frame f = (frame) alloc(sizeof(struct frame));
	  int i;

	  f->f_nvars = nvars;
	  f->f_next = NULL;
	  f->f_cxt = nil;
	  f->f_var = (type *) alloc(nvars * sizeof(type));
	  for (i = 0; i < nvars; i++)
	       f->f_var[i] = NULL;
	  return f;
     }
}

/* new_frame -- make a new frame and record it for checking */
PUBLIC frame new_frame(int nvars, tree cxt)
{
     frame f = mk_frame(nvars);

     if (f != arid) {
	  f->f_cxt = cxt;
	  f->f_next = stack[level].b_chain;
	  stack[level].b_chain = f;
     }
     
     return f;
}
